// g++ -Wall -Wpedantic -Wextra -std=c++17 -march=haswell -O3 auto-vec.cpp && ./a.out

#include <cstdint>
#include <random>
#include <algorithm>
#include <iostream>
#include <vector>
#include <numeric>
#include <cassert>
#include <x86intrin.h>
#include "vectorclass.h"

using namespace std;
#define LENGTH 12800000
#define REPS 100

//#define DATA_T int32_t
//#define RANDOM_DISTRIBUTION uniform_int_distribution
//#define DATA_PRINTF "%i"
#define DATA_T float
#define RANDOM_DISTRIBUTION uniform_real_distribution
#define DATA_PRINTF "%g"

DATA_T array_sum_1(DATA_T* array, uint64_t length) {
    DATA_T sum = 0;
    for (size_t i = 0; i < length; i++) {
        sum += array[i];
    }
    return sum;
}

__attribute__((optimize("-funsafe-math-optimizations")))
DATA_T array_sum_2(DATA_T* array, uint64_t length) {
    DATA_T sum = 0;
    for (size_t i = 0; i < length; i++) {
        sum += array[i];
    }
    return sum;
}

DATA_T kahan_sum(DATA_T* array, uint64_t length) {
    DATA_T sum = 0;
    DATA_T c = 0;
    DATA_T y, t;
    for (size_t i = 0; i < length; i++) {
        y = array[i] - c;
        t = sum + y;
        c = (t - sum) - y;
        sum = t;
    }
    return sum;
}


void time_array_sum(string name, DATA_T array_sum(DATA_T*, uint64_t), DATA_T* arr, uint64_t len) {
    DATA_T result = 0;
    uint64_t st, en;
    st = __rdtsc();
    for ( int i=0; i<REPS; i++ ) {
        result += array_sum(arr, len);
    }
    en = __rdtsc();
    printf("%s " DATA_PRINTF " in %g\n", name.c_str(), result, (double)(en-st)/(len*REPS));
}

int main(void) {
    // https://stackoverflow.com/a/23143753
    random_device rnd_device;
    mt19937 mersenne_engine {rnd_device()};
    RANDOM_DISTRIBUTION<DATA_T> dist {-10, 10};

    auto gen = [&dist, &mersenne_engine]() {
        return dist(mersenne_engine);
    };

    unsigned length = LENGTH;
    vector<DATA_T> vec(length);
    cout << "Array size: " << sizeof(DATA_T) * length / 1024 / 1024 << "MB\n";
    generate(begin(vec), end(vec), gen);

    DATA_T *array = &vec[0];
    time_array_sum("array_sum_1", array_sum_1, array, length);
    time_array_sum("array_sum_2", array_sum_2, array, length);
    time_array_sum("kahan_sum  ", kahan_sum, array, length);
    return 0;
}
